Module cryptors

Module cryptors 

Cryptor structs for encryption/decryption

A cryptor is a struct with methods that can encrypt (seal) a plaintext or decrypt (open) a ciphertext. This module provides RingCryptor, a cryptor struct that provides the above functionality using various ring cryptographic primitives.

Structs

RingCryptor
A cryptor that uses cryptographic primitives from the ring crate.